package com.winhc.data;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.*;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;

import cn.hutool.core.lang.Assert;
import com.winhc.data.task.BaseTask;
import com.winhc.data.task.SinkTask;
import com.winhc.data.task.DownLoadTask;
import com.winhc.data.utils.*;
import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.sql.DataSource;

import static com.winhc.data.utils.Common.*;
import static com.winhc.data.utils.DateUtils.getCurrentHours;
import static com.winhc.data.utils.EnvConst.initAllConf;
import static com.winhc.data.utils.EnvProperties.getYmlByFileNameOut;
import static com.winhc.data.utils.OssUtils.*;
import static java.lang.System.getenv;

/**
 * @author π
 * @Description:
 * @date 2023/5/29 11:26
 */
public class DataPull {
    private static final Logger logger = LoggerFactory.getLogger(DataPull.class);
    private static volatile boolean isRunning = false;

    public static void main(String[] args) throws Exception {

        Map<String, String> params = ParameterTool
                .fromArgs(args)
                .mergeWith(ParameterTool.fromSystemProperties())
                .mergeWith(ParameterTool.fromMap(getenv()))
                .toMap();
        String confPath;
        if (params.containsKey("docker.start")) {
            confPath = "/config.yml";
        } else {
            Assert.isTrue(params.containsKey("conf"), "请输入配置文件路径\t:\t--conf conf_path\n");
            confPath = params.get("conf");
        }

        Assert.isTrue(FileUtils.fileExists(confPath), "配置文件不存在\t:\t检查conf路径是否正确\n");
        Map<String, String> confMap = new HashMap<>();
        Assert.isTrue(getYmlByFileNameOut(confPath, confMap), "conf解析异常\t:\t配置文件必须yml格式\n");
        //merge params
        initAllConf(args, confMap);
        String env = EnvConst.getValue("profile.activate");
        logger.info("profile.activate: {}", env);
        Assert.isTrue(StringUtils.isNotBlank(env), "请输入环境配置egg:\nprofile:\n  activate: sj\n");

        int queueSize = 100000;
        int numThread = Integer.parseInt(EnvConst.getValue("winhc.task.numThread", "8"));
        int batchSize = Integer.parseInt(EnvConst.getValue("winhc.task.batchSize", "1000"));

        //单次运行
        //startOne(queueSize, numThread, batchSize);
        //保持进程
        startAllTime(queueSize, numThread, batchSize);

    }

    private static void startOne(int queueSize, int numThread, int batchSize) throws Exception {
        //获取任务
        while (true) {
            Boolean finish = exec(queueSize, numThread, batchSize);
            if (finish) break;
            //超过12点自动退出
            int currentHours = getCurrentHours();
            if (currentHours >= 18) {
                logger.info("waiting task time out , end hour {} !!!", currentHours);
                break;
            }
        }
    }

    private static void startAllTime(int queueSize, int numThread, int batchSize) throws Exception {
        //获取任务
        while (true) {
            exec(queueSize, numThread, batchSize);
        }
    }

    private static Boolean exec(int queueSize, int numThread, int batchSize) throws Exception {
        //获取下载文件
        List<String> uploadTasks = new ArrayList<>();
        CopyOnWriteArrayList<String> tasks = new CopyOnWriteArrayList<>(findTasks(uploadTasks));
        CopyOnWriteArrayList<String> success_task = new CopyOnWriteArrayList<>();

        long sleep = 1000 * 10;
        boolean flag = false;
        if (tasks.isEmpty()) {
            logger.info("no task to run !!!");
            Thread.sleep(sleep);
            return true;
        }
        for (String ds : tasks) {
            //判断文件是否上传
            if (existsFile(FILE_PRE + ds + FILE_UPLOAD_SUFFIX)) {
                startTask(numThread, batchSize, queueSize, ds);
                tasks.remove(ds);
                success_task.add(ds);
            }
            //剔除空数据
            if (uploadTasks.contains(ds)) {
                tasks.remove(ds);
            }
        }
        //休眠
        if (!tasks.isEmpty()) {
            logger.info("wait task \t{}\t!!!", String.join("\t|\t", tasks));
        } else {
            logger.info("task success \t{}\t!!!", String.join("\t|\t", success_task));
            flag = true;
        }
        Thread.sleep(sleep);
        return flag;
    }

    private static void startTask(int numThread, int batchNum, int queueSize, String ds) throws Exception {
        BlockingQueue<String> queue = new ArrayBlockingQueue<>(queueSize);
        ExecutorService executor = Executors.newFixedThreadPool(numThread + 1);
        DataSource dataSource = HikariUtil.getHikariCPDataSource();
        CountDownLatch latch = new CountDownLatch(numThread);
        List<SinkTask> tasks = new ArrayList<>();
        try {
            for (int k = 0; k < numThread; k++) {
                SinkTask workThread = new SinkTask(queue, dataSource.getConnection(), batchNum, latch, ds);
                tasks.add(workThread);
                executor.execute(workThread);
            }
        } catch (Exception e) {
            logger.error("add work thread error", e);
        }
        isRunning = true;
        Runnable statThread = () -> {
            while (isRunning) {
                long total = tasks.stream().map(SinkTask::getProcessNum).reduce(0L, Long::sum);
                logger.info("{} heartbeat {} thread process {}", ds, numThread, total);
                try {
                    TimeUnit.MILLISECONDS.sleep(2000);
                } catch (InterruptedException e) {
                    logger.error(e.getMessage(), e);
                    throw new RuntimeException(e.getMessage());
                }
            }
        };

        executor.execute(statThread);
        executor.shutdown();

        long start = System.currentTimeMillis();
        try {
            DownLoadTask downLoadTask = new DownLoadTask(queue, dataSource.getConnection(), ds);
            downLoadTask.downLoad(FILE_PRE + ds + FILE_UPLOAD_SUFFIX);
        } catch (Exception e) {
            success(numThread, queue);
            throw new RuntimeException(e.getMessage());
        }
        success(numThread, queue);
        latch.await();
        for (SinkTask task : tasks) {
            task.close();
        }
        executor.shutdownNow();
        long totalNum = tasks.stream().map(SinkTask::getProcessNum).reduce(0L, Long::sum);
        logger.info("{} all {} thread process finish | total count {} | total cost {} ms", ds, numThread, totalNum, (System.currentTimeMillis() - start));

        //拉取成功标志
        String sql2 = Common.UPDATE_SQL.replaceAll("@ds", ds).replaceAll("@status", "2");
        BaseTask baseTask = new BaseTask(dataSource.getConnection());
        baseTask.upsertStatus(sql2);
        baseTask.close();
        //上传状态
        saveStatus(ds, STATUS_SUCCESS_SUFFIX);
    }

    private static void success(int numThread, BlockingQueue<String> queue) {
        for (int i = 0; i < numThread; i++) {
            queue.add("QUIT");
        }
        isRunning = false;
    }

}
