package com.winhc.data.utils;

import com.zaxxer.hikari.HikariConfig;
import com.zaxxer.hikari.HikariDataSource;
import lombok.SneakyThrows;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;

import javax.sql.DataSource;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.sql.*;
import java.util.*;
import java.util.stream.Collectors;

/**
 * @author: XuJiakai
 * 2023/5/16 14:12
 * <p>
 * xjk  xjk_ABC123
 */

@Slf4j
public class HikariUtil {
    private final static HikariUtil HIKARI_CP_UTI = new HikariUtil();

    private static HikariDataSource dataSource = null;

    private HikariUtil() {
    }

    @SneakyThrows
    public static HikariUtil getInstance() {
        HIKARI_CP_UTI.registerHikariCP();
        return HIKARI_CP_UTI;

    }

    /**
     * 注册HikariCP
     */
    private synchronized void registerHikariCP() {
        if (null != dataSource) {
            return;
        }
        HikariConfig config = new HikariConfig();
        config.setDriverClassName(EnvConst.getValue("winhc.mysql.driverClassName"));
        config.setJdbcUrl(EnvConst.getValue("winhc.mysql.jdbcUrl"));
        config.setUsername(EnvConst.getValue("winhc.mysql.username"));
        config.setPassword(EnvConst.getValue("winhc.mysql.password"));
        int threadSize = Integer.parseInt(EnvConst.getValue("winhc.task.numThread", "20")) + 2;
        config.setMaximumPoolSize(threadSize * 2);
        config.setMinimumIdle(threadSize);
        config.setMaxLifetime(1800000);
        dataSource = new HikariDataSource(config);
    }

    /**
     * 提供对外 获取 HikariCPDatasource 的方法
     *
     * @return
     */
    public static DataSource getHikariCPDataSource() {
        if (null != dataSource) {
            return dataSource;
        }
        HIKARI_CP_UTI.registerHikariCP();
        return dataSource;
    }

    /**
     * 每次重新从连接池中获取新连接，注意：如连接不够会超时
     */
    public Connection getConn() throws SQLException {
        return getHikariCPDataSource().getConnection();
    }


    @SneakyThrows
    public static <T> List<T> exec(Connection conn, String sql, Class<T> typeClazz) {

        Statement statement = null;
        ResultSet rs = null;

        List<T> list = new ArrayList<>();
        Field[] declaredFields = typeClazz.getDeclaredFields();
        Method[] declaredMethods = typeClazz.getDeclaredMethods();
        Set<String> methodSet = Arrays.stream(declaredMethods).map(Method::getName).collect(Collectors.toSet());
        try {
            statement = conn.createStatement();
            rs = statement.executeQuery(sql);
            Set<String> fieldSet = getFieldSet(rs, typeClazz);

            while (rs.next()) {
                T instance = typeClazz.newInstance();
                for (Field field : declaredFields) {
                    Class<?> type = field.getType();
                    String fieldSetName = "set" + StringUtils.capitalize(field.getName());
                    if (!methodSet.contains(fieldSetName) || !fieldSet.contains(field.getName())) {
                        continue;
                    }
                    Method fieldSetMet = typeClazz.getMethod(fieldSetName, field
                            .getType());
                    switch (type.getSimpleName()) {
                        case "String":
                            fieldSetMet.invoke(instance, rs.getString(field.getName()));
                            break;
                        case "Integer":
                            fieldSetMet.invoke(instance, rs.getInt(field.getName()));
                            break;
                        default:
                            throw new RuntimeException();
                    }
                }
                list.add(instance);
            }
            return list;
        } catch (SQLException e) {
            log.error(e.getMessage(), e);
            log.error("hikari exec sql error: {}", e.getMessage());
            throw new RuntimeException("hikari exec sql error: " + e.getMessage(), e);
        } finally {
            if (rs != null) {
                rs.close();
            }
            if (statement != null) {
                statement.close();
            }
        }
    }


    private static final Map<Class<?>, Set<String>> map = new HashMap<>();

    private static Set<String> getFieldSet(ResultSet resultSet, Class<?> typeClazz) {
        if (!map.containsKey(typeClazz)) {
            synchronized (HikariUtil.class) {
                if (!map.containsKey(typeClazz)) {
                    try {
                        ResultSetMetaData metaData = resultSet.getMetaData();
                        int columnCount = metaData.getColumnCount();
                        HashSet<String> set = new HashSet<>();
                        for (int i = 1; i <= columnCount; i++) {
                            String columnName = metaData.getColumnName(i);
                            set.add(columnName);
                        }
                        map.put(typeClazz, set);
                    } catch (Exception e) {
                        log.error(e.getMessage(), e);
                        e.printStackTrace();
                        throw new RuntimeException(e);
                    }
                }
            }
        }
        return map.get(typeClazz);
    }
}
